/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2004-2005, Identity Alliance, LLC.                *
*  Title   : GTKCard.h                                                      *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

#define GTK_ENABLE_BROKEN 1	// Compatibility with GTK 1.2

#include <gtk/gtk.h>
#include <unix.h>
#include <pkcs11.h>
#include <List.h>
#include "pkcs11-display.h"

#define _P11Include 1

#define APP_ID        "GTKCard"
#define APP_VER       "0.0.1"
#define APP_TITLE     "GTKCard - Smart Card Management System"

enum
{
	TYPE_COLUMN,
	SIZE_COLUMN,
	INFORMATION_COLUMN,
	DATA_TYPE_COLUMN,
	DATA_POINTER_COLUMN,
	N_COLUMNS
};

enum
{
	TYPE_PIN = 1,
	TYPE_KEY,
	TYPE_CERTIFICATE,
	TYPE_DATA
};

// Global structures containing session/window variables
struct _globalState {
	CK_SESSION_HANDLE shSession;	// P11 session variable
	BOOL p11Initialized;	// P11 Initialized?
	GtkWidget *app;		// The main application window
	GtkTreeStore *treeStore;	// data model
	GtkWidget *mainTree;	// The main tree
	GtkWidget *tokenCombo;	// The combo box with cards
	GtkWidget *readerCombo;	// The combo box with readers
	GtkWidget *enrollCombo;	// The combo box with readers
	GtkWidget *profileCombo;	// Profile combo box
	GtkWidget *contentBox;	// Card contents textbox
	CK_BYTE selection;	// Which card is selected
	CK_BYTE readerSelection;	// Which reader is selected
	CK_BYTE enrollSelection;	// Which reader is selected
	CK_BYTE profileSelection;	// Which profile is selected
	GtkWidget *verify_button;
	GtkWidget *change_button;
	GtkWidget *view_button;
	GtkTreeIter objectIter;	// Object tree iterator
	GtkTreeIter keyIter;	// KEY tree iterator
	GtkTreeIter pinIter;	// PIN tree iterator
	BOOL quitStat;		// Quit printing star
	BOOL pause;		// Pause the * printing
	GdkFont *fixedFont;	// A fixed font for display purpose
	GtkWidget *executeProfileButton;
	GtkWidget *refreshProfilesButton;
	GtkEntry *hostEntry;
	GtkEntry *portEntry;
};

struct _textBoxInfo {
	gchar *text;		// Data object in text format
	short textLen;		// Length of data object
	GtkWidget *tmpText;	// Textbox to display data object
	BOOL hex;		// Are we printing HEX or ASCII
};

struct _pinItemInfo {
	GtkWidget *entry;	// PIN entry
	BOOL startup;		// PIN verification on startup?
	BOOL userType;		// CKU_USER or CKU_SO
};

struct _globalState globalState;
struct _textBoxInfo textBoxInfo;
struct _pinItemInfo pinItemInfo;

// A global variable for the textBox - *** Used for overloading the macros ***
extern GtkWidget *textBox;

extern gchar fileName[FILENAME_MAX];

// Global constants
#define FILE_NAME_LEN            256
#define RSA_KEY_SIZE             128

// Debug definition, comment out for no debug messages
#ifdef USE_DEBUG
#define DEBUG                    printf
#else
#define DEBUG
#endif

// Assertions and printing macros
#define ASSERT(b, msg, rv) if (!b) { DELETE(); \
                                     display_main("%s\tError: %s (0x%08X)\n", \
						  msg, lookup_enum(RV_T, rv), rv); \
                                     return FALSE; }
#define DISPLAY(msg) gtk_text_insert(GTK_TEXT(textBox), NULL, \
                                     &textBox->style->black, NULL, msg, -1)
#define DISPLAY_FIXED(msg) gtk_text_insert(GTK_TEXT(textBox), \
                                           globalState.fixedFont, \
                                           &textBox->style->black, NULL, \
                                           msg, -1)
#define DISPLAY_RED(msg) { GdkColor color; color.red = 0xffff; \
                           color.green = 0; color.blue = 0; \
                           gtk_text_insert(GTK_TEXT(textBox), NULL, \
                                           &color, NULL, msg, -1); }
#define DISPLAY_BLUE(msg) { GdkColor color; color.red = 0; \
                            color.green = 0; color.blue = 0xffff ; \
                            gtk_text_insert(GTK_TEXT(textBox), \
                                           globalState.fixedFont, \
                                           &color, NULL, msg, -1); }
#define DELETE() gtk_text_backward_delete(GTK_TEXT(textBox), \
                                        gtk_text_get_length(GTK_TEXT(textBox)))
#define FLUSH() while (gtk_events_pending()) \
                  gtk_main_iteration_do(0);

// Utility structures
typedef struct changePin {
	GtkEntry *oldPIN;
	GtkEntry *newPIN1;
	GtkEntry *newPIN2;
} changePinInfo, *pChangePinInfo;

typedef struct enrollUser {
	GtkEntry *userName;
	GtkEntry *fileName;
} enrollUserInfo, *pEnrollUserInfo;

// Implemented functions

// Window functions
BOOL createWindows();
void createMainMenu(GtkWidget * window, GtkWidget ** menubar);
void printStatStars();
void createObjectWindow(pObjectInfo object);
void createCertWindow(pCertInfo cert);
void createP12Dialog();
void createCertDialog();
void clearTree();

// Event callback functions
void change_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data);
void delete_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data);
void connect_event_cb(GtkWidget * window, gpointer data);
void disconnect_event_cb(GtkWidget * window, gpointer data);
void pinclick_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data);
void pinok_event_cb(GtkWidget * window, gpointer data);
void change_ok_event_cb(GtkWidget * window, gpointer data);
void dialog_close_event_cb(GtkWidget * window, gpointer data);
void object_close_event_cb(GtkWidget * window, gpointer data);
void file_ok_cb(GtkWidget * w, GtkFileSelection * fs);
void cancel_file_cb(GtkWidget * widget, gpointer data);
void cert_ok_event_cb(GtkWidget * window, gpointer data);
void file_p12_ok_cb(GtkWidget * w, GtkFileSelection * fs);
void keyclick_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data);
void certclick_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data);
void objectclick_event_cb(GtkWidget * window, GdkEventButton * e,
			  gpointer data);
void file_obj_ok_cb(GtkWidget * w, GtkFileSelection * fs);
void ascii_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data);
void hex_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data);
void file_dialog_close_event_cb(GtkWidget * window, gpointer data);
void p12_ok_event_cb(GtkWidget * window, gpointer data);
void refresh_event_cb(GtkWidget * window, gpointer data);
void ma_connect_event_cb(GtkWidget * window, gpointer data);
void ma_execute_event_cb(GtkWidget * window, gpointer data);
void ma_refreshReaders_event_cb(GtkWidget * window, gpointer data);
void ma_refreshProfiles_event_cb(GtkWidget * window, gpointer data);
void reader_change_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data);
void profile_selection_made_cb(GtkWidget * window,
			       GdkEventAny * e, gpointer data);
void ma_destroy_event_cb(GtkWidget * window, gpointer data);
void enroll_change_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data);
void enroll_ok_event_cb(GtkWidget * window, gpointer data);
void enroll_close_event_cb(GtkWidget * window, gpointer data);
void enroll_destroy_event_cb(GtkWidget * window, gpointer data);
void view_button_cb(GtkWidget * window, gpointer data);

// P#11 functions
BOOL listTokens(BOOL reset);
BOOL openSession();
BOOL closeSession();
BOOL readObjects(GtkTreeIter * objectParent, GtkTreeIter * keyParent);
BOOL login(CK_USER_TYPE userType, const char *pinCode);
BOOL changePIN(const char *pinCode, const char *oldPIN);
BOOL importCert(gchar * fileName, const char * certID);
BOOL importP12(gchar * fileName, const char * passwd);
BOOL readKeyInfo(CK_OBJECT_HANDLE ohObject, CK_OBJECT_CLASS ocClass);
BOOL readCertInfo(CK_OBJECT_HANDLE ohObject);
BOOL readObjectInfo(CK_OBJECT_HANDLE ohObject);
BOOL createDataObject(gchar * fileName);
BOOL P11_ReadObject(CK_OBJECT_HANDLE ohObject, gchar ** text);
BOOL P11_displayCert(pCertInfo certInfo, GtkWidget * textBox);

// Menu Functions
void m_verifyPIN(GtkWidget * w, gpointer data);
void m_ObjectsDelete(GtkWidget * w, gpointer data);

// Display messages from the 
void display(char *msg, ...);
void display_main(char *msg, ...);
void messageBox(char *msg, ...);
