/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2004-2005, Identity Alliance, LLC.                *
*  Title   : List.h                                                         *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

typedef short BOOL;

enum listTypes {
	KEY,
	OBJECT,
	CERT
};

#define NAME_LEN        128
#define CERT_NAME_LEN   96

typedef struct keyStruct {
	CK_UTF8CHAR name[NAME_LEN];
	CK_OBJECT_CLASS class;
	CK_UTF8CHAR keyId[NAME_LEN];
	CK_KEY_TYPE keyType;
	CK_OBJECT_HANDLE ohHandle;
} keyInfo, *pKeyInfo;

typedef struct objectStruct {
	CK_UTF8CHAR name[NAME_LEN];
	CK_UTF8CHAR oId[NAME_LEN];
	CK_ULONG size;
	CK_OBJECT_HANDLE ohHandle;
} objectInfo, *pObjectInfo;

typedef struct certObject {
	CK_UTF8CHAR name[NAME_LEN];
	CK_UTF8CHAR keyId[NAME_LEN];
	CK_UTF8CHAR issuer[CERT_NAME_LEN];
	CK_UTF8CHAR subject[CERT_NAME_LEN];
	CK_UTF8CHAR serialNumber[CERT_NAME_LEN];
	CK_ULONG size;
	CK_OBJECT_HANDLE ohHandle;
} certInfo, *pCertInfo;

typedef struct pinStruct {
	CK_UTF8CHAR name[NAME_LEN];
	BOOL verified;
} pinInfo, *pPinInfo;

typedef struct listContainer {
	union elements {
		pKeyInfo key;
		pObjectInfo object;
		pCertInfo cert;
	} element;
	struct listContainer *next;
} container, *pContainer;

typedef struct listHead {
	CK_BYTE type;
	pContainer first;
} list, *pList;

struct _lists {
	list keys;
	list objects;
	list certs;
};

struct _pins {
	pinInfo adminPin;
	pinInfo userPin;
};

struct _lists lists;
struct _pins pins;

BOOL append(pList l, CK_ULONG_PTR item);
BOOL freeList(pList l);
BOOL initLists();
BOOL length(pList l, CK_ULONG_PTR len);
