/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2003-2005, Identity Alliance, LLC.                *
*  Title   : CreateWindows.c                                                *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <unistd.h>

#include "GTKCard.h"
#include "Mutex.h"

// A global variable for the textBox - *** Used for overloading the macros ***
GtkWidget *textBox;

// An external mutex variable used for the pin dialog
pthread_mutex_t pinLock;

static void m_changePIN(GtkWidget * w, gpointer data);
static void m_importCert(GtkWidget * w, gpointer data);
static void m_importPKCS12(GtkWidget * w, gpointer data);
static void m_about(GtkWidget * w, gpointer data);
static void m_ObjectsCreate(GtkWidget * w, gpointer data);
static void m_manageApplets(GtkWidget * w, gpointer data);
static void m_enrollUser(GtkWidget * w, gpointer data);
static void m_quit(GtkWidget * w, gpointer data);

static GtkItemFactoryEntry mainMenuItems[] = {
	{"/_File", NULL, NULL, 0, "<Branch>"},
	{"/File/_Manage Applets", NULL, m_manageApplets, 0, NULL},
	{"/File/_Quit", "<control>Q", m_quit, 0, "<StockItem>", GTK_STOCK_QUIT},
	{"/_Objects", NULL, NULL, 0, "<Branch>"},
	{"/Objects/_Create Object", NULL, m_ObjectsCreate, 0, "<StockItem>", GTK_STOCK_ADD},
	{"/Objects/_Delete Object", NULL, m_ObjectsDelete, 0, "<StockItem>", GTK_STOCK_REMOVE},
	{"/_Certs", NULL, NULL, 0, "<Branch>"},
	{"/Certs/Import _Cert (DER)", NULL, m_importCert, 0, NULL},
	{"/Certs/Import _PKCS#12 Cert", NULL, m_importPKCS12, 0, NULL},
	//  { "/_Keys",                    NULL,          NULL, 0, "<Branch>" },
	//  { "/Keys/_Generate Keys",      NULL,          m_genKeys, 0, NULL },
	{"/_Users", NULL, NULL, 0, "<Branch>"},
	{"/Users/_Enroll users", NULL, m_enrollUser, 0, NULL},
	{"/_Help", NULL, NULL, 0, "<LastBranch>"},
	{"/_Help/_About", NULL, m_about, 0, "<StockItem>", GTK_STOCK_ABOUT},
};

// Global function implementations

BOOL createWindows()
{
	// Widgets to be used in the window
	GtkWidget *tokenLabel;
	GtkWidget *refreshButton;
	GtkWidget *connectButton;
	GtkWidget *disconnectButton;
	GtkWidget *containerTable;
	GtkWidget *menuBar;
	GtkWidget *scrollWindow;
	GtkWidget *vbox, *hbox;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;

	// Assign the graphical variables with new components
	globalState.app = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(globalState.app), APP_TITLE);

	vbox = gtk_vbox_new(FALSE, 0);
	containerTable = gtk_table_new(24, 36, TRUE);
	tokenLabel = gtk_label_new("Please select token:");
	gtk_label_set_justify(GTK_LABEL(tokenLabel), GTK_JUSTIFY_LEFT);
	connectButton = gtk_button_new_with_label("Connect");
	disconnectButton = gtk_button_new_with_label("Disconnect");
	globalState.tokenCombo = gtk_combo_new();
	globalState.treeStore = gtk_tree_store_new (N_COLUMNS,
			G_TYPE_STRING,	/* Type */
			G_TYPE_UINT,	/* Size */
			G_TYPE_STRING,	/* Information */
			G_TYPE_CHAR,	/* object type */
			G_TYPE_POINTER);	/* data object pointer */
	globalState.mainTree = gtk_tree_view_new_with_model(GTK_TREE_MODEL(globalState.treeStore));
	renderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("Type", renderer,
			"text", TYPE_COLUMN, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (globalState.mainTree), column);
	column = gtk_tree_view_column_new_with_attributes ("Size", renderer,
			"text", SIZE_COLUMN, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (globalState.mainTree), column);
	column = gtk_tree_view_column_new_with_attributes ("Information", renderer,
			"text", INFORMATION_COLUMN, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (globalState.mainTree), column);
	refreshButton = gtk_button_new_with_label("Refresh");

	// Set the window porperties
	gtk_window_set_policy(GTK_WINDOW(globalState.app), TRUE, TRUE, TRUE);
	gtk_window_set_wmclass(GTK_WINDOW(globalState.app), APP_ID, APP_TITLE);

	// Take care of the buttons
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 6);
	//gtk_container_set_border_width(GTK_CONTAINER(connectButton), 10);
	//gtk_container_set_border_width(GTK_CONTAINER(disconnectButton), 10);
	//gtk_container_set_border_width(GTK_CONTAINER(refreshButton), 10);

	// Create a scroll window for the tree
	scrollWindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrollWindow),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	// Add the tree to the scrolled window
	//gtk_tree_set_selection_mode(GTK_TREE(globalState.mainTree),
	//	GTK_SELECTION_SINGLE);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrollWindow),
		globalState.mainTree);

	// Create the textBox
	textBox = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(textBox), FALSE);

	// Assign elements to the container table
	gtk_misc_set_alignment(GTK_MISC(tokenLabel), 0, 1);
	gtk_table_attach(GTK_TABLE(containerTable), tokenLabel,
		0, 17, 0, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), globalState.tokenCombo,
		0, 17, 2, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), connectButton,
		0, 5, 4, 6, GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), disconnectButton,
		6, 11, 4, 6, GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), refreshButton,
		12, 17, 4, 6, GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), textBox,
		18, 36, 0, 6,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), scrollWindow,
		0, 36, 7, 24,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		GTK_FILL | GTK_EXPAND | GTK_SHRINK, 0, 0);

	gtk_entry_set_editable(GTK_ENTRY
		(GTK_COMBO(globalState.tokenCombo)->entry), FALSE);
	globalState.selection = 1;	// The default selected token
	globalState.shSession = 0;
	globalState.fixedFont = gdk_font_load("fixed");

	// Assign the contents of the application
	gtk_container_add(GTK_CONTAINER(globalState.app), vbox);

	// Now for the main menu
	createMainMenu(globalState.app, &menuBar);
	gtk_box_pack_start(GTK_BOX(vbox), menuBar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), containerTable, FALSE, FALSE, 0);

	// button bar
	hbox = gtk_hbox_new(TRUE, 6);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 6);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	globalState.verify_button = gtk_button_new_with_label("Verify PIN");
	gtk_box_pack_start(GTK_BOX(hbox), globalState.verify_button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(globalState.verify_button), "clicked",
		GTK_SIGNAL_FUNC(m_verifyPIN), NULL);
	gtk_widget_set_sensitive(globalState.verify_button, FALSE);

	globalState.change_button = gtk_button_new_with_label("Change PIN");
	gtk_box_pack_start(GTK_BOX(hbox), globalState.change_button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(globalState.change_button), "clicked",
		GTK_SIGNAL_FUNC(m_changePIN), NULL);
	gtk_widget_set_sensitive(globalState.change_button, FALSE);

	globalState.view_button = gtk_button_new_with_label("View");
	gtk_box_pack_start(GTK_BOX(hbox), globalState.view_button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(globalState.view_button), "clicked",
		GTK_SIGNAL_FUNC(view_button_cb), NULL);
	gtk_widget_set_sensitive(globalState.view_button, FALSE);

	// Associate functions with events
	gtk_signal_connect(GTK_OBJECT(globalState.app),
		"delete_event", GTK_SIGNAL_FUNC(delete_event_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(globalState.tokenCombo)->entry),
		"changed", GTK_SIGNAL_FUNC(change_event_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(connectButton),
		"clicked", GTK_SIGNAL_FUNC(connect_event_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(refreshButton),
		"clicked", GTK_SIGNAL_FUNC(refresh_event_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(disconnectButton),
		"clicked", GTK_SIGNAL_FUNC(disconnect_event_cb), NULL);

	gtk_widget_show_all(globalState.app);

	// Now list the tokens using P#11
	globalState.p11Initialized = FALSE;
	listTokens(TRUE);

	if (globalState.p11Initialized)
	{
		CK_RV rv;
		CK_ULONG ulSlotCount = 0;

		ASSERT((rv = (C_GetSlotList(FALSE, NULL_PTR, &ulSlotCount))) == CKR_OK,
			"Could not get slot list", rv);

		/* only one slot it present so we connect directly */
		if (1 == ulSlotCount)
			gtk_button_clicked(GTK_BUTTON(connectButton));
	}

	return TRUE;
}

static void m_manageApplets(GtkWidget * w, gpointer data)
{
	GtkWidget *manageAppletWindow;
	GtkWidget *comboLabel;
	GtkWidget *profileLabel;
	GtkWidget *contentLabel;
	GtkWidget *connectButton;
	GtkWidget *refreshReadersButton;
	GtkWidget *containerTable;
	GtkWidget *vScrollBar;
	GtkWidget *hostLabel;
	GtkWidget *portLabel;

	// List the tokens using the IACSDK first
	globalState.readerCombo = gtk_combo_new();

	// Labels and Buttons
	comboLabel = gtk_label_new("Please select smartcard reader:");
	gtk_label_set_justify(GTK_LABEL(comboLabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(comboLabel), 0, 1);
	profileLabel = gtk_label_new("Please select profile:");
	gtk_label_set_justify(GTK_LABEL(profileLabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(profileLabel), 0, 1);
	contentLabel = gtk_label_new("Card Contents:");
	gtk_label_set_justify(GTK_LABEL(contentLabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(contentLabel), 0, 1);
	connectButton = gtk_button_new_with_label("Connect");
	refreshReadersButton = gtk_button_new_with_label("Refresh");
	globalState.executeProfileButton = gtk_button_new_with_label("Execute");
	gtk_widget_set_sensitive(globalState.executeProfileButton, FALSE);
	globalState.refreshProfilesButton = gtk_button_new_with_label("Refresh");
	gtk_widget_set_sensitive(globalState.refreshProfilesButton, FALSE);
	hostLabel = gtk_label_new("Applet loader server:");
	gtk_label_set_justify(GTK_LABEL(hostLabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(hostLabel), 0, 1);
	globalState.hostEntry = GTK_ENTRY(gtk_entry_new());
	gtk_entry_set_text(globalState.hostEntry, "home.identityalliance.com");
	portLabel = gtk_label_new("Server port:");
	gtk_label_set_justify(GTK_LABEL(portLabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(portLabel), 0, 1);
	globalState.portEntry = GTK_ENTRY(gtk_entry_new());
	gtk_entry_set_max_length(GTK_ENTRY(globalState.portEntry), 5);
	gtk_entry_set_text(GTK_ENTRY(globalState.portEntry), "22");

	// Create the textBox
	globalState.contentBox = gtk_text_new(NULL, NULL);
	vScrollBar = gtk_vscrollbar_new(GTK_TEXT(globalState.contentBox)->vadj);
	gtk_text_set_editable(GTK_TEXT(globalState.contentBox), FALSE);

	// Create the listBox
	globalState.profileCombo = gtk_combo_new();

	// The window and the container table
	manageAppletWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	containerTable = gtk_table_new(12, 36, TRUE);
	gtk_window_set_policy(GTK_WINDOW(manageAppletWindow), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(manageAppletWindow), "Manage Applets");
	gtk_window_set_position(GTK_WINDOW(manageAppletWindow),
		GTK_WIN_POS_MOUSE);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 10);

	// Assign elements to the container table
	gtk_table_attach(GTK_TABLE(containerTable), hostLabel,
		0, 17, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable),
		GTK_WIDGET(globalState.hostEntry), 0, 17, 1, 2, GTK_FILL | GTK_EXPAND,
		GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), portLabel, 18, 36, 0, 1,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable),
		GTK_WIDGET(globalState.portEntry), 18, 36, 1, 2,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), comboLabel, 0, 17, 2, 3,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), globalState.readerCombo, 0,
		17, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), connectButton, 0, 5, 4, 5,
		GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), refreshReadersButton, 6, 11,
		4, 5, GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), profileLabel, 18, 36, 2, 3,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable),
		globalState.executeProfileButton, 18, 23, 4, 5, GTK_FILL | GTK_EXPAND,
		GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable),
		globalState.refreshProfilesButton, 24, 29, 4, 5,
		GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), globalState.profileCombo, 18,
		36, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), contentLabel, 0, 34, 5, 6,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), globalState.contentBox, 0, 34,
		6, 12, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), vScrollBar, 34, 36, 6, 12,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);

	gtk_container_add(GTK_CONTAINER(manageAppletWindow), containerTable);

	globalState.readerSelection = 1;	// The default selected reader
	globalState.profileSelection = 1;	// The default selected profile

	// Associate functions with events
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(globalState.readerCombo)->entry),
		"changed", GTK_SIGNAL_FUNC(reader_change_event_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(globalState.profileCombo)->entry),
		"changed", GTK_SIGNAL_FUNC(profile_selection_made_cb), NULL);
	gtk_signal_connect(GTK_OBJECT(connectButton), "clicked",
		GTK_SIGNAL_FUNC(ma_connect_event_cb), NULL);

	gtk_widget_show_all(manageAppletWindow);
}

static void m_changePIN(GtkWidget * w, gpointer data)
{
	// Create a window to ask for the PIN
	GtkWidget *changePinWindow;
	GtkWidget *okButton;
	GtkWidget *cancelButton;
	GtkWidget *oldEntry;
	GtkWidget *newEntry1;
	GtkWidget *newEntry2;
	GtkWidget *oldLabel;
	GtkWidget *newLabel1;
	GtkWidget *newLabel2;
	GtkWidget *containerTable;
	GtkWidget *vbox1, *hbox1;

	pChangePinInfo changeInfo = malloc(sizeof(changePinInfo));

	// Labels and buttons
	oldLabel = gtk_label_new("Old PIN:");
	newLabel1 = gtk_label_new("New PIN:");
	newLabel2 = gtk_label_new("Repeat:");
	okButton = gtk_button_new_from_stock(GTK_STOCK_OK);
	cancelButton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	oldEntry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(oldEntry), 16);
	gtk_entry_set_visibility(GTK_ENTRY(oldEntry), FALSE);
	newEntry1 = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(newEntry1), 16);
	gtk_entry_set_visibility(GTK_ENTRY(newEntry1), FALSE);
	newEntry2 = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(newEntry2), 16);
	gtk_entry_set_visibility(GTK_ENTRY(newEntry2), FALSE);

	// The window and the container table
	changePinWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	containerTable = gtk_table_new(3, 2, FALSE);
	gtk_window_set_title(GTK_WINDOW(changePinWindow), "Change PIN");
	gtk_window_set_position(GTK_WINDOW(changePinWindow), GTK_WIN_POS_MOUSE);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 12);

	vbox1 = gtk_vbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(changePinWindow), vbox1);

	hbox1 = gtk_hbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(vbox1), hbox1);

	gtk_box_pack_start(GTK_BOX(hbox1),
		gtk_image_new_from_stock(GTK_STOCK_DIALOG_AUTHENTICATION,
		GTK_ICON_SIZE_DIALOG), FALSE, FALSE, 12);

	// Assign elements to the container
	gtk_misc_set_alignment(GTK_MISC(oldLabel), 0, 0);
	gtk_misc_set_alignment(GTK_MISC(newLabel1), 0, 0);
	gtk_misc_set_alignment(GTK_MISC(newLabel2), 0, 0);

	gtk_table_attach(GTK_TABLE(containerTable), oldLabel,
		0, 1, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 5, 0);
	gtk_table_attach(GTK_TABLE(containerTable), newLabel1,
		0, 1, 1, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 5, 0);
	gtk_table_attach(GTK_TABLE(containerTable), newLabel2,
		0, 1, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 5, 0);
	gtk_table_attach(GTK_TABLE(containerTable), oldEntry,
		1, 2, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), newEntry1,
		1, 2, 1, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), newEntry2,
		1, 2, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_box_pack_start(GTK_BOX(hbox1), containerTable, FALSE, FALSE, 2);

	// buttons
	hbox1 = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox1), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbox1), 6);
	gtk_container_set_border_width(GTK_CONTAINER(hbox1), 12);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 2);

	gtk_box_pack_start(GTK_BOX(hbox1), cancelButton, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox1), okButton, FALSE, FALSE, 0);

	// Assign events
	changeInfo->oldPIN = GTK_ENTRY(oldEntry);
	changeInfo->newPIN1 = GTK_ENTRY(newEntry1);
	changeInfo->newPIN2 = GTK_ENTRY(newEntry2);
	DEBUG("ChangeInfo: %p\n", changeInfo);

	gtk_signal_connect(GTK_OBJECT(okButton),
		"clicked", GTK_SIGNAL_FUNC(change_ok_event_cb), changeInfo);
	gtk_signal_connect(GTK_OBJECT(cancelButton),
		"clicked", GTK_SIGNAL_FUNC(dialog_close_event_cb), NULL);

	gtk_widget_show_all(changePinWindow);
}

void m_verifyPIN(GtkWidget * w, gpointer data)
{
	// Create a window to ask for the PIN
	GtkWidget *pinWindow;
	GtkWidget *okButton;
	GtkWidget *cancelButton;
	GtkWidget *entry;
	GtkWidget *label;
	GtkWidget *vbox1, *hbox1;

	// Labels and buttons
	label = gtk_label_new("Please enter PIN code:");
	okButton = gtk_button_new_from_stock(GTK_STOCK_OK);
	cancelButton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	entry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(entry), 16);
	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);

	// The window
	pinWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(pinWindow), "Verify PIN");
	gtk_window_set_position(GTK_WINDOW(pinWindow), GTK_WIN_POS_MOUSE);

	vbox1 = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox1), label, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox1), entry, FALSE, FALSE, 5);

	hbox1 = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox1),
		gtk_image_new_from_stock(GTK_STOCK_DIALOG_AUTHENTICATION,
		GTK_ICON_SIZE_DIALOG), FALSE, FALSE, 12);
	gtk_box_pack_start(GTK_BOX(hbox1), vbox1, FALSE, FALSE, 12);

	vbox1 = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(pinWindow), vbox1);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

	hbox1 = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox1), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbox1), 6);
	gtk_container_set_border_width(GTK_CONTAINER(hbox1), 12);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 2);

	gtk_box_pack_start(GTK_BOX(hbox1), cancelButton, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox1), okButton, FALSE, FALSE, 0);

	// Set the pinInfo variables
	pinItemInfo.entry = entry;

	// Assign events
	gtk_signal_connect(GTK_OBJECT(okButton),
		"clicked", GTK_SIGNAL_FUNC(pinok_event_cb), &pinItemInfo);
	gtk_signal_connect(GTK_OBJECT(cancelButton),
		"clicked", GTK_SIGNAL_FUNC(dialog_close_event_cb), &pinItemInfo);

	gtk_widget_show_all(pinWindow);

	if (pinItemInfo.startup)
	{
		// Want to lock the mutex
		DEBUG("CreateWindow locks the mutex\n");
		lockMutex(&pinLock);
		globalState.pause = TRUE;	// Stop the printing of *'s
	}
}

static void m_importCert(GtkWidget * w, gpointer data)
{
	GtkWidget *fileWindow;
	GtkFileFilter *filter;

	if (globalState.shSession == NULL_PTR)
	{
		DELETE();
		DISPLAY_RED("ERROR: You must connect to a card first");
		return;
	}

	if (!pins.userPin.verified)
	{
		DELETE();
		DISPLAY_RED("ERROR: PIN code is not verified.");
		return;
	}

	DELETE();
	DISPLAY("Importing Certificate, please wait...");

	fileWindow = gtk_file_chooser_dialog_new("Select a .der certificate File",
		NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, ".der files");
	gtk_file_filter_add_pattern(filter, "*.der");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fileWindow), filter);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "all files");
	gtk_file_filter_add_pattern(filter, "*");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fileWindow), filter);

	if (gtk_dialog_run(GTK_DIALOG(fileWindow)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename;

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileWindow));
		g_strlcpy(fileName, filename, sizeof(fileName));
		g_free(filename);

		createCertDialog();
	}

	gtk_widget_destroy(fileWindow);
}

static void m_importPKCS12(GtkWidget * w, gpointer data)
{
	GtkWidget *fileWindow;
	GtkFileFilter *filter;

	if (globalState.shSession == NULL_PTR)
	{
		DELETE();
		DISPLAY_RED("ERROR: You must connect to a card first");
		return;
	}

	if (!pins.userPin.verified)
	{
		DELETE();
		DISPLAY_RED("ERROR: PIN code is not verified.");
		return;
	}

	if (!pins.userPin.verified)
	{
		DEBUG("Not verified\n");
		DISPLAY("Cannot import certificate, PIN not verified.\n");
		return;
	}

	DEBUG("Verified\n");

	DELETE();
	DISPLAY("Importing Certificate, please wait...");

	fileWindow = gtk_file_chooser_dialog_new("Select a .p12 certificate File",
		NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, ".p12 files");
	gtk_file_filter_add_pattern(filter, "*.p12");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fileWindow), filter);

	filter = gtk_file_filter_new();
	gtk_file_filter_set_name(filter, "all files");
	gtk_file_filter_add_pattern(filter, "*");
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fileWindow), filter);

	if (gtk_dialog_run(GTK_DIALOG(fileWindow)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename;

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileWindow));
		g_strlcpy(fileName, filename, sizeof(fileName));
		g_free(filename);

		createP12Dialog();
	}

	gtk_widget_destroy(fileWindow);
}

static void m_ObjectsCreate(GtkWidget * w, gpointer data)
{
	GtkWidget *fileWindow;

	if (globalState.shSession == NULL_PTR)
	{
		DELETE();
		DISPLAY_RED("ERROR: You must connect to a card first");
		return;
	}

	DELETE();
	DISPLAY("Creating object, please wait...");

	fileWindow = gtk_file_chooser_dialog_new("Select a File",
		NULL, GTK_FILE_CHOOSER_ACTION_OPEN,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	if (gtk_dialog_run(GTK_DIALOG(fileWindow)) == GTK_RESPONSE_ACCEPT)
	{
		char *filename;

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fileWindow));
		createDataObject(filename);
		g_free(filename);
	}

	gtk_widget_destroy(fileWindow);
}

#if 0
static void m_genKeys(GtkWidget * w, gpointer data)
{
	CK_MECHANISM mechanism;
	CK_RV rv;
	CK_OBJECT_HANDLE hPublicKey, hPrivateKey;
	CK_ULONG modulusBits = RSA_KEY_SIZE * 8;
	CK_BYTE publicExponent[3];
	CK_BBOOL bTrue = TRUE;
	CK_ATTRIBUTE publicKeyTemplate[] = {
		{CKA_ENCRYPT, &bTrue, sizeof(bTrue)}
		,
		{CKA_VERIFY, &bTrue, sizeof(bTrue)}
		,
		{CKA_TOKEN, &bTrue, sizeof(bTrue)}
		,
		{CKA_WRAP, &bTrue, sizeof(bTrue)}
		,
		{CKA_MODULUS_BITS, &modulusBits, sizeof(modulusBits)}
		,
		{CKA_PUBLIC_EXPONENT, publicExponent, sizeof(publicExponent)}
	};
	CK_ATTRIBUTE privateKeyTemplate[] = {
		{CKA_TOKEN, &bTrue, sizeof(bTrue)}
		,
		{CKA_PRIVATE, &bTrue, sizeof(bTrue)}
		,
		{CKA_TOKEN, &bTrue, sizeof(bTrue)}
		,
		{CKA_SENSITIVE, &bTrue, sizeof(bTrue)}
		,
		{CKA_DECRYPT, &bTrue, sizeof(bTrue)}
		,
		{CKA_SIGN, &bTrue, sizeof(bTrue)}
		,
		{CKA_UNWRAP, &bTrue, sizeof(bTrue)}
	};
	pthread_t pthThread;

	if (globalState.shSession == NULL_PTR)
	{
		DELETE();
		DISPLAY_RED("ERROR: You must connect to a card first");
		return;
	}

	mechanism.pParameter = NULL;
	mechanism.ulParameterLen = 0;
	mechanism.mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN;

	// Set the public exponent - hex for 65537
	publicExponent[0] = 0x01;
	publicExponent[1] = 0x00;
	publicExponent[2] = 0x01;

	DELETE();
	DISPLAY("Generating keys [");
	FLUSH();

	globalState.quitStat = FALSE;
	rv = pthread_create(&pthThread, NULL, (void *) &printStatStars, 0);

	rv = C_GenerateKeyPair(globalState.shSession,
		&mechanism,
		publicKeyTemplate,
		6, privateKeyTemplate, 7, &hPublicKey, &hPrivateKey);

	if (rv != CKR_OK)
	{
		globalState.quitStat = TRUE;
		while (globalState.quitStat == TRUE)
			usleep(1000 * 10);
		DEBUG("Return: %lx\n", rv);
		DISPLAY_RED("ERROR: Could not generate key pair\n");
		return;
	}

	readKeyInfo(hPublicKey, CKO_PUBLIC_KEY);
	readKeyInfo(hPrivateKey, CKO_PRIVATE_KEY);

	globalState.quitStat = TRUE;
	while (globalState.quitStat == TRUE)
		usleep(1000 * 10);

	DISPLAY("\nKey Generation complete");
}
#endif

static void m_enrollUser(GtkWidget * w, gpointer data)
{
	GtkWidget *enrollUserWindow;
	GtkWidget *comboLabel;
	GtkWidget *containerTable;
	GtkWidget *okButton;
	GtkWidget *cancelButton;
	GtkWidget *fileEntry;
	GtkWidget *fileLabel;
	GtkWidget *userEntry;
	GtkWidget *userLabel;
	GtkWidget *label;

	pEnrollUserInfo enrollInfo = malloc(sizeof(enrollUserInfo));

	// Labels, Buttons and Combos
	label = gtk_label_new("NOTE: You need to do this as root!");
	comboLabel = gtk_label_new("Please select smartcard reader:");
	gtk_label_set_justify(GTK_LABEL(comboLabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(comboLabel), 0, 1);
	okButton = gtk_button_new_from_stock(GTK_STOCK_OK);
	cancelButton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);

	fileLabel = gtk_label_new("User file:");
	gtk_misc_set_alignment(GTK_MISC(fileLabel), 0, 0.5);
	fileEntry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(fileEntry), 25);
	gtk_entry_set_text(GTK_ENTRY(fileEntry), "/etc/pam_smartcard.users");
	userLabel = gtk_label_new("User Name:");
	gtk_misc_set_alignment(GTK_MISC(userLabel), 0, 0.5);
	userEntry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(userEntry), 16);

	globalState.enrollCombo = gtk_combo_new();

	enrollInfo->userName = (GtkEntry *) userEntry;
	enrollInfo->fileName = (GtkEntry *) fileEntry;

	// The window and the container table
	enrollUserWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	containerTable = gtk_table_new(7, 3, TRUE);
	gtk_window_set_policy(GTK_WINDOW(enrollUserWindow), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(enrollUserWindow), "Enroll User");
	gtk_window_set_position(GTK_WINDOW(enrollUserWindow), GTK_WIN_POS_MOUSE);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 10);
	gtk_container_set_border_width(GTK_CONTAINER(okButton), 10);
	gtk_container_set_border_width(GTK_CONTAINER(cancelButton), 10);

	// Assign elements to the container
	gtk_table_attach(GTK_TABLE(containerTable), label,
		0, 3, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), comboLabel,
		0, 3, 1, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), globalState.enrollCombo,
		0, 3, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), fileLabel,
		0, 1, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), fileEntry,
		1, 3, 3, 4, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), userLabel,
		0, 1, 4, 5, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), userEntry,
		1, 3, 4, 5, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), okButton,
		0, 1, 5, 7, GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), cancelButton,
		2, 3, 5, 7, GTK_FILL | GTK_EXPAND, GTK_SHRINK | GTK_EXPAND, 0, 0);

	gtk_container_add(GTK_CONTAINER(enrollUserWindow), containerTable);

	gtk_widget_show_all(enrollUserWindow);
}

static void m_quit(GtkWidget * w, gpointer data)
{
	DEBUG("Releasing memory, user clicked quit\n");
	closeSession();
	freeList(&(lists.keys));
	freeList(&(lists.objects));
	freeList(&(lists.certs));
	free(textBoxInfo.text);

	gtk_main_quit();
}

static void m_about(GtkWidget * w, gpointer data)
{
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new(GTK_WINDOW(globalState.app),
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_INFO,
		GTK_BUTTONS_CLOSE,
		"GTKCard - A Smart Card Management System\n\nCreated by Identity Alliance for Sun Microsystems\n\nWritten by Eirik A. Herskedal\nVersion 0.1.0");

	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

void createCertDialog()
{
	// Create a window to ask for the PIN
	GtkWidget *certDialog;
	GtkWidget *okButton;
	GtkWidget *cancelButton;
	GtkWidget *entry;
	GtkWidget *label;
	GtkWidget *containerTable;

	// Labels and buttons
	label = gtk_label_new("Enter private key number\nto associate with certificate:");
	okButton = gtk_button_new_from_stock(GTK_STOCK_OK);
	cancelButton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	entry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(entry), 3);

	// The window and the container table
	certDialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	containerTable = gtk_table_new(5, 2, TRUE);
	gtk_window_set_policy(GTK_WINDOW(certDialog), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(certDialog), "Select private Key");
	gtk_window_set_position(GTK_WINDOW(certDialog), GTK_WIN_POS_MOUSE);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 10);
	gtk_container_set_border_width(GTK_CONTAINER(okButton), 10);
	gtk_container_set_border_width(GTK_CONTAINER(cancelButton), 10);

	// Assign elements to the container
	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0);
	gtk_table_attach(GTK_TABLE(containerTable), label,
		0, 2, 0, 2, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), entry,
		0, 2, 2, 3, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), okButton,
		0, 1, 3, 5, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), cancelButton,
		1, 2, 3, 5, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);

	gtk_container_add(GTK_CONTAINER(certDialog), containerTable);

	// Assign events
	gtk_signal_connect(GTK_OBJECT(okButton),
		"clicked", GTK_SIGNAL_FUNC(cert_ok_event_cb), entry);
	gtk_signal_connect_object(GTK_OBJECT(cancelButton),
		"clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(certDialog));

	gtk_widget_show_all(certDialog);
}

void createP12Dialog()
{
	// Create a window to ask for the PIN
	GtkWidget *p12Dialog;
	GtkWidget *okButton;
	GtkWidget *cancelButton;
	GtkWidget *entry;
	GtkWidget *label;
	GtkWidget *vbox1, *hbox1;

	// Labels and buttons
	label = gtk_label_new("Please enter password:");
	okButton = gtk_button_new_from_stock(GTK_STOCK_OK);
	cancelButton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);

	// The window
	p12Dialog = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(p12Dialog), "Enter password");
	gtk_window_set_position(GTK_WINDOW(p12Dialog), GTK_WIN_POS_MOUSE);

	vbox1 = gtk_vbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox1), label, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(vbox1), entry, FALSE, FALSE, 5);

	hbox1 = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox1),
		gtk_image_new_from_stock(GTK_STOCK_DIALOG_AUTHENTICATION,
		GTK_ICON_SIZE_DIALOG), FALSE, FALSE, 12);
	gtk_box_pack_start(GTK_BOX(hbox1), vbox1, FALSE, FALSE, 12);

	vbox1 = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(p12Dialog), vbox1);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

	hbox1 = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox1), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(hbox1), 6);
	gtk_container_set_border_width(GTK_CONTAINER(hbox1), 12);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 2);

	gtk_box_pack_start(GTK_BOX(hbox1), cancelButton, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox1), okButton, FALSE, FALSE, 0);

	// Assign events
	gtk_signal_connect(GTK_OBJECT(okButton),
		"clicked", GTK_SIGNAL_FUNC(p12_ok_event_cb), entry);
	gtk_signal_connect_object(GTK_OBJECT(cancelButton),
		"clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(p12Dialog));

	gtk_widget_show_all(p12Dialog);
}

void createCertWindow(pCertInfo cert)
{
	// Create a window to display a certificate
	GtkWidget *objectWindow;
	GtkWidget *closeButton;
	GtkWidget *textBox;
	GtkWidget *containerTable;
	GtkWidget *vscrollbar;
	GtkWidget *vbox, *hbox;

	// Labels and buttons
	closeButton = gtk_button_new_from_stock(GTK_STOCK_CLOSE);

	// Create the textBox and a scrollbar
	textBox = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(textBox), FALSE);
	vscrollbar = gtk_vscrollbar_new(GTK_TEXT(textBox)->vadj);

	// The window and the container table
	objectWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	containerTable = gtk_table_new(12, 24, TRUE);
	gtk_window_set_policy(GTK_WINDOW(objectWindow), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(objectWindow), "View Certificate");
	gtk_window_set_position(GTK_WINDOW(objectWindow), GTK_WIN_POS_MOUSE);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 15);
	//  gtk_container_set_border_width(GTK_CONTAINER(closeButton), 10);

	gtk_table_attach(GTK_TABLE(containerTable), textBox,
		0, 23, 0, 12, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), vscrollbar,
		23, 24, 0, 12,
		GTK_FILL | GTK_EXPAND, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(objectWindow), vbox);
	gtk_box_pack_start(GTK_BOX(vbox), containerTable, TRUE, TRUE, 0);

	hbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 12);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(hbox), closeButton, FALSE, FALSE, 0);

	if (!P11_displayCert(cert, textBox))
	{
		DISPLAY_RED("ERROR: Could not display certificate\n");
	}
	// Assign events
	gtk_signal_connect(GTK_OBJECT(closeButton), "clicked",
			GTK_SIGNAL_FUNC(dialog_close_event_cb), NULL);

	gtk_widget_show_all(objectWindow);
}

void createObjectWindow(pObjectInfo object)
{
	// Create a window to display an object
	GtkWidget *objectWindow;
	GtkWidget *closeButton;
	GtkWidget *textBox;
	GtkWidget *hexButton;
	GtkWidget *asciiButton;
	GtkWidget *vscrollbar;
	GtkWidget *vbox, *hbox;
	GSList *group;

	// Labels and buttons
	closeButton = gtk_button_new_from_stock(GTK_STOCK_CLOSE);

	// Create the textBox
	textBox = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(textBox), FALSE);
	vscrollbar = gtk_vscrollbar_new(GTK_TEXT(textBox)->vadj);

	// The window and the container table
	objectWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_policy(GTK_WINDOW(objectWindow), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(objectWindow), "View Data Object");
	gtk_window_set_position(GTK_WINDOW(objectWindow), GTK_WIN_POS_MOUSE);

	// Now for the radio buttons
	hexButton = gtk_radio_button_new_with_label(NULL, "Hex");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(hexButton));
	asciiButton = gtk_radio_button_new_with_label(group, "ASCII");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(hexButton), TRUE);

	vbox = gtk_vbox_new(FALSE, 12);
	gtk_container_add(GTK_CONTAINER(objectWindow), vbox);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), textBox, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), vscrollbar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

	hbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	gtk_box_pack_start(GTK_BOX(hbox), hexButton, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), asciiButton, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), closeButton, FALSE, FALSE, 0);

	textBoxInfo.text = NULL;
	textBoxInfo.textLen = object->size;
	textBoxInfo.hex = TRUE;
	if (P11_ReadObject(object->ohHandle, &(textBoxInfo.text)))
	{
		int i;
		int j = 16;
		gchar string[5];
		DELETE();
		snprintf(string, 5, "%02X: ", 0);
		DISPLAY_BLUE(string);
		for (i = 0; i < object->size; i++)
		{
			snprintf(string, 3, "%02X", textBoxInfo.text[i]);
			DISPLAY_FIXED(string);
			if ((i + 1) % 16 == 0)
			{
				DISPLAY("\n");
				snprintf(string, 5, "%02X: ", j);
				DISPLAY_BLUE(string);
				j += 16;
			}
			else
				DISPLAY(" ");
		}
	}
	else
	   	if (textBoxInfo.text != NULL)
			DISPLAY_RED("ERROR: Could not read object from card");

	// Assign events
	textBoxInfo.tmpText = textBox;
	DEBUG("TextBoxInfo: %p\n", &textBoxInfo);

	gtk_signal_connect(GTK_OBJECT(hexButton),
		"clicked", GTK_SIGNAL_FUNC(hex_event_cb), NULL);

	gtk_signal_connect(GTK_OBJECT(asciiButton),
		"clicked", GTK_SIGNAL_FUNC(ascii_event_cb), NULL);

	gtk_signal_connect(GTK_OBJECT(closeButton), "clicked",
		GTK_SIGNAL_FUNC(object_close_event_cb), NULL);

	gtk_widget_show_all(objectWindow);
}

void createMainMenu(GtkWidget * window, GtkWidget ** menubar)
{
	GtkItemFactory *itemFactory;
	GtkAccelGroup *accelGroup;
	gint nmenu_items = sizeof(mainMenuItems) / sizeof(mainMenuItems[0]);

	accelGroup = (GtkAccelGroup *) gtk_accel_group_new();

	itemFactory = GTK_ITEM_FACTORY(gtk_item_factory_new(GTK_TYPE_MENU_BAR,
		"<main>", accelGroup));

	gtk_item_factory_create_items(itemFactory, nmenu_items, mainMenuItems,
		NULL);

	gtk_window_add_accel_group(GTK_WINDOW(window), accelGroup);

	if (menubar)
		*menubar = gtk_item_factory_get_widget(itemFactory, "<main>");
}

void printStatStars()
{
	while (TRUE)
	{
		if (!globalState.pause)
		{
			usleep(500000);
			DISPLAY("*");
		}
		else
			usleep(100000);
		FLUSH();
		if (globalState.quitStat)
			break;
	}

	DISPLAY("] : ");
	FLUSH();
	globalState.quitStat = FALSE;
	globalState.pause = FALSE;
}

void display(char *msg, ...)
{
	char string[256];
	va_list args;
	GtkWidget *textBox = globalState.contentBox;

	va_start(args, msg);

	vsnprintf(string, sizeof(string), msg, args);

	DISPLAY(string);

	va_end(args);
}

void display_main(char *msg, ...)
{
	char string[256];
	va_list args;

	va_start(args, msg);

	vsnprintf(string, sizeof(string), msg, args);

	DISPLAY(string);

	va_end(args);
}

void messageBox(char *msg, ...)
{
	// Create a window to display a message
	GtkWidget *msgWindow;
	GtkWidget *okButton;
	//  GtkWidget *cancelButton;
	GtkWidget *label;
	GtkWidget *containerTable;

	char string[256];
	va_list args;

	va_start(args, msg);

	vsnprintf(string, sizeof(string), msg, args);

	va_end(args);

	// Labels and buttons
	label = gtk_label_new(string);
	okButton = gtk_button_new_from_stock(GTK_STOCK_OK);
	//  cancelButton = gtk_button_new_with_label("Cancel");

	// The window and the container table
	msgWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	containerTable = gtk_table_new(3, 3, TRUE);
	gtk_window_set_policy(GTK_WINDOW(msgWindow), TRUE, TRUE, TRUE);
	gtk_window_set_title(GTK_WINDOW(msgWindow), "Error");

	gtk_window_set_position(GTK_WINDOW(msgWindow), GTK_WIN_POS_MOUSE);

	// Take care of the borders
	gtk_container_set_border_width(GTK_CONTAINER(containerTable), 10);
	gtk_container_set_border_width(GTK_CONTAINER(okButton), 10);
	//  gtk_container_set_border_width(GTK_CONTAINER(cancelButton), 10);

	// Assign elements to the container
	//  gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), label,
		0, 3, 0, 1, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	gtk_table_attach(GTK_TABLE(containerTable), okButton,
		1, 2, 1, 3, GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
	//  gtk_table_attach(GTK_TABLE(containerTable), cancelButton,
	//                   1, 2, 2, 4,
	//                   GTK_FILL | GTK_EXPAND,
	//                   GTK_FILL | GTK_EXPAND,
	//                   0, 0);

	gtk_container_add(GTK_CONTAINER(msgWindow), containerTable);

	// Assign events
	gtk_signal_connect(GTK_OBJECT(okButton),
		"clicked", GTK_SIGNAL_FUNC(dialog_close_event_cb), NULL);
	//  gtk_signal_connect(GTK_OBJECT(cancelButton),
	//                     "clicked",
	//                     GTK_SIGNAL_FUNC(dialog_close_event_cb),
	//                     &pinItemInfo);

	gtk_widget_show_all(msgWindow);
}

void clearTree()
{
	gtk_tree_store_clear(globalState.treeStore);

	freeList(&(lists.keys));
	freeList(&(lists.objects));
	freeList(&(lists.certs));

	//  globalState.mainTree = gtk_tree_new();
}

