/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2003-2005, Identity Alliance, LLC.                *
*  Title   : EventHandlers.c                                                *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>

#include <GTKCard.h>
#include <Mutex.h>

#include "pkcs11-display.h"

// External variables
GtkWidget *textBox;
pthread_mutex_t pinLock;
gchar fileName[FILENAME_MAX];

// Global variables for this file
static struct _view_object
{
	char type;
	gpointer data;
} ViewObject;

// Quit the application
void delete_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data)
{
	DEBUG("Releasing memory\n");
	closeSession();
	freeList(&(lists.keys));
	freeList(&(lists.objects));
	freeList(&(lists.certs));
	free(textBoxInfo.text);

	gtk_main_quit();
}

// The token comboBox changes
void change_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data)
{
	globalState.selection = atoi(gtk_entry_get_text(GTK_ENTRY(window)));
	DEBUG("Selection: %d\n", globalState.selection);
}

static void tree_selection_changed_cb(GtkTreeSelection *selection,
	gpointer data)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	gpointer ptr;
	gchar type;

	if (!gtk_tree_selection_get_selected(selection, &model, &iter))
	{
		DEBUG("Data model not found!\n");
		return;
	}

	gtk_tree_model_get(model, &iter, DATA_TYPE_COLUMN, &type,
		DATA_POINTER_COLUMN, &ptr, -1);

	switch (type)
	{
		case TYPE_PIN:
			{
				gchar *label;
				pPinInfo pin;

				gtk_tree_model_get(model, &iter, TYPE_COLUMN, &label, -1);

				DELETE();
				if (label[0] == 'A')
				{
					DISPLAY("Administrator PIN\n");
					pin = &(pins.adminPin);
					pinItemInfo.userType = CKU_SO;
				}
				else
				{
					DISPLAY("User PIN\n");
					pin = &(pins.userPin);
					pinItemInfo.userType = CKU_USER;
				}

				if (pin->verified)
					DISPLAY("Verified");
				else
					DISPLAY("Not verified");

				gtk_widget_set_sensitive(globalState.verify_button, TRUE);
				gtk_widget_set_sensitive(globalState.change_button, TRUE);
				gtk_widget_set_sensitive(globalState.view_button, FALSE);
			}
			break;

		case TYPE_KEY:
			{
				int i;
				gchar string[200];
				pKeyInfo key = (pKeyInfo) ptr;

				snprintf(string, 20, "Key, ");

				DELETE();
				DISPLAY(string);
				if (key->class == CKO_PRIVATE_KEY)
					DISPLAY("RSA Private\n");
				else
					DISPLAY("RSA Public\n");

				DISPLAY("Key ID:\n");
				DISPLAY_FIXED(" ");
				for (i = 0; i < 20; i++)
				{
					snprintf(string, 3, "%02X", key->keyId[i]);
					DISPLAY_FIXED(string);
					if (i != 0 && i != 18 && (i % 9 == 0))
						DISPLAY_FIXED("\n ");
					else
						DISPLAY(" ");
				}
				DISPLAY("\n");

				ViewObject.type = TYPE_KEY;
				ViewObject.data = key;

				gtk_widget_set_sensitive(globalState.verify_button, FALSE);
				gtk_widget_set_sensitive(globalState.change_button, FALSE);
				gtk_widget_set_sensitive(globalState.view_button, FALSE);
			}
			break;

		case TYPE_CERTIFICATE:
			{
				int i;
				gchar string[200];
				pCertInfo cert = (pCertInfo) ptr;

				snprintf(string, 20, "Certificate\n");

				DELETE();
				DISPLAY(string);

				DISPLAY("ID:\n");
				DISPLAY_FIXED(" ");
				for (i = 0; i < 20; i++)
				{
					snprintf(string, 4, "%02X", cert->keyId[i]);
					DISPLAY_FIXED(string);
					if (i != 0 && i != 18 && (i % 9 == 0))
						DISPLAY_FIXED("\n ");
					else if (i != 19)
						DISPLAY_FIXED(" ");
				}
				DISPLAY("\n");

				snprintf(string, sizeof(string), "Size: %ld", cert->size);
				DISPLAY(string);

				ViewObject.type = TYPE_CERTIFICATE;
				ViewObject.data = cert;

				gtk_widget_set_sensitive(globalState.verify_button, FALSE);
				gtk_widget_set_sensitive(globalState.change_button, FALSE);
				gtk_widget_set_sensitive(globalState.view_button, TRUE);
			}
			break;

		case TYPE_DATA:
			{
				gchar string[200];
				pObjectInfo object = (pObjectInfo) ptr;

				DELETE();
				DISPLAY("Object\n");

				snprintf(string, sizeof(string), "Size: %ld\n", object->size);
				DISPLAY(string);

				ViewObject.type = TYPE_DATA;
				ViewObject.data = object;

				gtk_widget_set_sensitive(globalState.verify_button, FALSE);
				gtk_widget_set_sensitive(globalState.change_button, FALSE);
				gtk_widget_set_sensitive(globalState.view_button, TRUE);
			}
			break;

		case 0:
			/* tree father */
			gtk_widget_set_sensitive(globalState.verify_button, FALSE);
			gtk_widget_set_sensitive(globalState.change_button, FALSE);
			gtk_widget_set_sensitive(globalState.view_button, FALSE);
			break;

		default:
			DEBUG("Unknown type: %d\n", type);
	}
}


// The use clicked connect - need to build the tree with information
void connect_event_cb(GtkWidget * window, gpointer data)
{
	pthread_t pthThread;
	GtkTreeIter iter;
	GtkTreeSelection *select;

	if (globalState.shSession == NULL_PTR)
	{
		DELETE();
		DISPLAY("Connecting [");
	}
	else
	{
		clearTree();
		//closeSession();
		DELETE();
		DISPLAY("Refreshing [");
	}
	FLUSH();

	globalState.quitStat = FALSE;
	pthread_create(&pthThread, NULL, (void *) &printStatStars, 0);

	if (globalState.shSession == NULL_PTR)
	{
		if (!openSession())
		{
			globalState.quitStat = TRUE;
			while (globalState.quitStat == TRUE)
				usleep(1000 * 10);

			DISPLAY("Could not connect");

			return;
		}
	}

	// Create the parent items in the tree
	gtk_tree_store_append(globalState.treeStore, &globalState.objectIter, NULL);
	gtk_tree_store_set(globalState.treeStore, &globalState.objectIter, TYPE_COLUMN, "Objects", -1);

	gtk_tree_store_append(globalState.treeStore, &globalState.keyIter, NULL);
	gtk_tree_store_set(globalState.treeStore, &globalState.keyIter, TYPE_COLUMN, "Keys", -1);

	if (!readObjects(&globalState.objectIter, &globalState.keyIter))
	{
		globalState.quitStat = TRUE;
		while (globalState.quitStat == TRUE)
			usleep(1000 * 10);

		DISPLAY("Could not connect");
		return;
	}

	// Create the PIN items
	gtk_tree_store_append(globalState.treeStore, &globalState.pinIter, NULL);
	gtk_tree_store_set(globalState.treeStore, &globalState.pinIter, TYPE_COLUMN, "PINs", -1);

	snprintf((char *) pins.adminPin.name, sizeof(pins.adminPin.name), "%s",
		"Admin PIN");
	pins.adminPin.verified = FALSE;
	gtk_tree_store_append(globalState.treeStore, &iter, &globalState.pinIter);
	gtk_tree_store_set(globalState.treeStore, &iter, TYPE_COLUMN,
		pins.adminPin.name, INFORMATION_COLUMN, "Not verified",
	   	DATA_TYPE_COLUMN, TYPE_PIN, DATA_POINTER_COLUMN, &pins.adminPin, -1);

	snprintf((char *) pins.userPin.name, sizeof(pins.userPin.name), "%s",
		"User PIN");
	//pins.userPin.verified = FALSE; Possibly verified on startup
	gtk_tree_store_append(globalState.treeStore, &iter, &globalState.pinIter);
	gtk_tree_store_set(globalState.treeStore, &iter, TYPE_COLUMN,
		pins.userPin.name, INFORMATION_COLUMN,
	   	pins.userPin.verified ? "Verified" : "Not verified",
		DATA_TYPE_COLUMN, TYPE_PIN, DATA_POINTER_COLUMN, &pins.userPin, -1);

	select = gtk_tree_view_get_selection(GTK_TREE_VIEW(globalState.mainTree));
	gtk_tree_selection_set_mode(select, GTK_SELECTION_SINGLE);
	g_signal_connect(G_OBJECT(select), "changed",
		G_CALLBACK(tree_selection_changed_cb), NULL);

	// 2 PINs
	gtk_tree_store_set(globalState.treeStore, &globalState.pinIter,
		SIZE_COLUMN, 2, -1);

	// Display the modified tree
	gtk_widget_show_all(globalState.app);

	globalState.quitStat = TRUE;
	while (globalState.quitStat == TRUE)
		usleep(1000 * 10);

	DISPLAY("\nInformation read from token");
}

// The use clicked disconnect
void disconnect_event_cb(GtkWidget * window, gpointer data)
{
	closeSession();
	clearTree();
}

// The user clicked refresh, list tokens again
void refresh_event_cb(GtkWidget * window, gpointer data)
{
	listTokens(FALSE);
}

// Ready to verify PIN
void pinok_event_cb(GtkWidget * window, gpointer data)
{
	struct _pinItemInfo *pinItemInfo = (struct _pinItemInfo *) data;
	GtkEntry *entry = GTK_ENTRY(pinItemInfo->entry);
	BOOL rv;

	if (!pinItemInfo->startup)
	{
		CK_USER_TYPE userType;

		userType = pinItemInfo->userType;
		DEBUG("UserType: %ld\tCKU_SO: %d\tCKU_USER: %d\n", userType,
			CKU_SO, CKU_USER);

		DELETE();
		rv = login(userType, gtk_entry_get_text(entry));
		if (rv)
			DISPLAY("PIN Verified Successfully");
		else
			DISPLAY("Could not verify PIN");

		if (userType == CKU_SO)
			pins.adminPin.verified = rv;
		else
			pins.userPin.verified = rv;

		// refresh the objects list
		connect_event_cb(NULL, 0);
	}
	else
	{
		rv = login(CKU_USER, gtk_entry_get_text(entry));

		if (rv)
			pins.userPin.verified = TRUE;

		DEBUG("Eventhandler unlocks mutex, rv = %x\n", rv);
		unlockMutex(&pinLock);
	}

	gdk_window_destroy((GdkWindow *) gtk_widget_get_parent_window(window));
}

// Close the PIN popup window
void dialog_close_event_cb(GtkWidget * window, gpointer data)
{
	gdk_window_destroy((GdkWindow *) gtk_widget_get_parent_window(window));
	if (data != NULL_PTR && ((struct _pinItemInfo *) data)->startup)
	{
		DEBUG("Eventhandler unlocks mutex, PIN_CANCEL\n");
		unlockMutex(&pinLock);
	}
}

// Close the Object popup window
void object_close_event_cb(GtkWidget * window, gpointer data)
{
	free(textBoxInfo.text);
	gdk_window_destroy((GdkWindow *) gtk_widget_get_parent_window(window));
}

// Ready to change the PIN
void change_ok_event_cb(GtkWidget * window, gpointer data)
{
	pChangePinInfo changeInfo = (pChangePinInfo) data;
	GtkEntry *oldEntry = changeInfo->oldPIN;
	GtkEntry *newEntry1 = changeInfo->newPIN1;
	GtkEntry *newEntry2 = changeInfo->newPIN2;

	DEBUG("Data: %p\n", data);
	DELETE();
	if (strncmp(gtk_entry_get_text(newEntry1),
			gtk_entry_get_text(newEntry2), 16))
	{
		DISPLAY("New PINs do not match");
		free(changeInfo);
		return;
	}

	if (changePIN(gtk_entry_get_text(oldEntry), gtk_entry_get_text(newEntry1)))
		DISPLAY("PIN Changed Successfully");
	else
		DISPLAY("Could not change the PIN");

	free(changeInfo);
	gdk_window_destroy((GdkWindow *) gtk_widget_get_parent_window(window));
}

// The user want hex
void hex_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data)
{
	GtkWidget *textBox = textBoxInfo.tmpText;
	int i;
	int j = 16;
	gchar string[5];

	if (textBoxInfo.hex)
		return;

	DELETE();
	snprintf(string, 5, "%02X: ", 0);
	DISPLAY_BLUE(string);
	for (i = 0; i < textBoxInfo.textLen; i++)
	{
		snprintf(string, 3, "%02X", textBoxInfo.text[i]);
		DISPLAY_FIXED(string);
		if ((i + 1) % 16 == 0)
		{
			DISPLAY("\n");
			snprintf(string, 5, "%02X: ", j);
			DISPLAY_BLUE(string);
			j += 16;
		}
		else
			DISPLAY(" ");
	}

	textBoxInfo.hex = TRUE;
}

// The user want ascii
void ascii_event_cb(GtkWidget * window, GdkEventButton * e, gpointer data)
{
	GtkWidget *textBox = textBoxInfo.tmpText;
	int i;
	gchar string[3];

	if (!textBoxInfo.hex)
		return;

	DELETE();
	for (i = 0; i < textBoxInfo.textLen; i++)
	{
		if (isascii(textBoxInfo.text[i]) || isspace(textBoxInfo.text[i]))
			g_snprintf(string, sizeof(string), "%c", textBoxInfo.text[i]);
		else
			g_snprintf(string, sizeof(string), ".");
		DISPLAY(string);
	}

	textBoxInfo.hex = FALSE;
}

// Ready to import the certificate
void cert_ok_event_cb(GtkWidget * window, gpointer data)
{
	GtkEntry *entry = GTK_ENTRY(data);

	DELETE();
	if (importCert(fileName, gtk_entry_get_text(entry)))
		DISPLAY("Certificate Imported Successfully");
	else
		DISPLAY("Could not import certificate");

	gdk_window_destroy((GdkWindow *) gtk_widget_get_parent_window(window));
}

// Ready to import the P12 file
void p12_ok_event_cb(GtkWidget * window, gpointer data)
{
	GtkEntry *entry = GTK_ENTRY(data);

	DELETE();
	if (importP12(fileName, gtk_entry_get_text(entry)))
		DISPLAY("P12 File Imported Successfully");
	else
		DISPLAY("Could not import P12 file");

	gdk_window_destroy((GdkWindow *) gtk_widget_get_parent_window(window));
}

// The reader comboBox changes
void reader_change_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data)
{
	globalState.readerSelection = atoi(gtk_entry_get_text(GTK_ENTRY(window)));
	DEBUG("Reader selection: %d\n", globalState.readerSelection);
}

// The user selected a profile
void profile_selection_made_cb(GtkWidget * window, GdkEventAny * e,
	gpointer data)
{
	globalState.profileSelection = atoi(gtk_entry_get_text(GTK_ENTRY(window)));

	DEBUG("Profile selection: %d\n", globalState.profileSelection);
}

// The user wants to connect to GP card
void ma_connect_event_cb(GtkWidget * window, gpointer data)
{
	GtkWidget *tmp = textBox;
	textBox = globalState.contentBox;

	DELETE();
	DISPLAY("Error: Could not connect to the specified reader.");
	FLUSH();

	textBox = tmp;
}

// The token comboBox changes
void enroll_change_event_cb(GtkWidget * window, GdkEventAny * e, gpointer data)
{
	globalState.enrollSelection = atoi(gtk_entry_get_text(GTK_ENTRY(window)));
}

void view_button_cb(GtkWidget * window, gpointer data)
{
	if (TYPE_CERTIFICATE == ViewObject.type)
		createCertWindow((pCertInfo)ViewObject.data);

	if (TYPE_DATA == ViewObject.type)
		createObjectWindow((pObjectInfo)ViewObject.data);
}

void m_ObjectsDelete(GtkWidget * w, gpointer data)
{
	CK_RV rv;
	CK_OBJECT_HANDLE hObject;

	switch (ViewObject.type)
	{
		case TYPE_CERTIFICATE:
			hObject = ((pCertInfo)ViewObject.data)->ohHandle;
			break;
		case TYPE_DATA:
			hObject = ((pObjectInfo)ViewObject.data)->ohHandle;
			break;
		case TYPE_KEY:
			hObject = ((pKeyInfo)ViewObject.data)->ohHandle;
			break;
		default:
			DEBUG("Unknown type: %d\n", ViewObject.type);
			return;
	}

	rv = C_DestroyObject(globalState.shSession, hObject);

	DELETE();
	if (CKR_OK == rv)
	{
		// refresh the objects list
		connect_event_cb(NULL, 0);

		DISPLAY("Deletion suceeded");
	}
	else
	{
		DELETE();
		display_main("Could not delete the object: %s (0x%08lX)",
			lookup_enum(RV_T, rv), rv);
	}
} /* m_ObjectsDelete */

