/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2003-2005, Identity Alliance, LLC.                *
*  Title   : Lists.c                                                        *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

#include <stdlib.h>

#include <GTKCard.h>

BOOL initLists()
{
	lists.keys.type = KEY;
	lists.objects.type = OBJECT;
	lists.certs.type = CERT;
	if (lists.keys.first != NULL)
		freeList(&(lists.keys));
	if (lists.objects.first != NULL)
		freeList(&(lists.objects));
	if (lists.certs.first != NULL)
		freeList(&(lists.certs));
	return TRUE;
}

BOOL freeList(pList l)
{
	pContainer prev, next;
	if (l == NULL_PTR)
		return TRUE;
	prev = l->first;
	while (prev != NULL) {
		next = prev->next;
		free(prev);
		prev = next;
	}
	l->first = NULL_PTR;
	return TRUE;
}

BOOL append(pList l, CK_ULONG_PTR item)
{
	pContainer c = l->first;
	if (c == NULL) {
		c = malloc(sizeof(container));
		ASSERT(c != NULL_PTR, "Could not allocate memory", -1);
		l->first = c;
		c->next = NULL;
	} else {
		while (c->next != NULL)
			c = c->next;
		c->next = malloc(sizeof(container));
		ASSERT(c != NULL_PTR, "Could not allocate memory", -1);
		c = c->next;
		c->next = NULL;
	}
	switch (l->type) {
	case KEY:
		c->element.key = (pKeyInfo) item;
		break;
	case OBJECT:
		c->element.object = (pObjectInfo) item;
		break;
	case CERT:
		c->element.cert = (pCertInfo) item;
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

BOOL length(pList l, CK_ULONG_PTR len)
{
	pContainer tmp = l->first;
	*len = 0;

	while (tmp != NULL) {
		(*len)++;
		tmp = tmp->next;
	}

	return TRUE;
}
