/****************************************************************************
*                                                                           *
*                                                                           *
*  GTKCard, Copyright (C) 2003-2005, Identity Alliance, LLC.                *
*  Title   : MuscleFormat.c                                                 *
*  Package : GTKCard                                                        *
*  Author  : Eirik A. Herskedal                                             *
*            David Corcoran                                                 *
*  Date    : 09/03/03                                                       *
*  Licence : See attached file                                              *
*                                                                           *
*                                                                           *
****************************************************************************/

#include <stdlib.h>

#include <GTKCard.h>
#include <Mutex.h>
#include <ctype.h>

#ifdef MUSCLE

// External variables
GtkWidget *textBox;

BOOL textToBytes(char *str, unsigned char *buf, unsigned long *bufLen)
{
	int i;
	int j;
	int len;

	j = 0;
	len = 0;

	len = strlen(str);

	if (len > 16) {
		messageBox("Transport key too long, max length 16.");
		return FALSE;
	}

	if (len <= 0) {
		messageBox("Transport too short.");
		return FALSE;
	}

	if (len % 2 != 0) {
		messageBox("Error: Transport key length not divisible by 2.");
		return FALSE;
	}

	for (i = 0; i < len; i += 2) {
		if (str[i] <= '9' && str[i] >= '0') {
			buf[j] = (str[i] - '0') * 16;
		} else if (str[i] <= 'F' && str[i] >= 'A') {
			buf[j] = (str[i] - 'A' + 10) * 16;
		} else {
			messageBox
			    ("Error parsing the transport key at char %d.", i);
			return FALSE;
		}

		if (str[i + 1] <= '9' && str[i + 1] >= '0') {
			buf[j] += str[i + 1] - '0';
		} else if (str[i + 1] <= 'F' && str[i + 1] >= 'A') {
			buf[j] += str[i + 1] - 'A' + 10;
		} else {
			messageBox
			    ("Error parsing the transport key at char %d.",
			     i + 1);
			return FALSE;
		}

		j += 1;
	}

	*bufLen = j;

	return TRUE;
}

BOOL MSCFillCombo()
{
	MSCLong32 rv;
	IACChar8 infoString[256];
	int i;

	// GTK variables
	GList *tokenItems = NULL_PTR;

	rv = MSCListTokens(MSC_LIST_KNOWN, NULL, &(globalState.tokenSize));
	if (rv != MSC_SUCCESS) {
		DELETE();
		DISPLAY_RED("ERROR: ListTokens Failed");
		return FALSE;
	}

	if (globalState.tokenSize == 0) {
		DELETE();
		DISPLAY_RED("ERROR: No Valid Tokens Found\n");
		return FALSE;
	}

	globalState.tokenList = malloc(sizeof(MSCTokenInfo) *
		globalState.tokenSize);

	//  DEBUG("TokenSize: %d\tTokenList addr: %x\n", globalState.tokenSize,
	//    globalState.tokenList);

	rv = MSCListTokens(MSC_LIST_KNOWN, globalState.tokenList,
			   &(globalState.tokenSize));

	if (rv != MSC_SUCCESS) {
		DELETE();
		DISPLAY_RED("ERROR: ListTokens Failed");
		return FALSE;
	}

	for (i = 0; i < globalState.tokenSize; i++) {
		snprintf(infoString, 256, "%d: %s", i + 1,
			 globalState.tokenList[i].tokenName);
		tokenItems = g_list_append(tokenItems, strdup(infoString));
	}

	gtk_combo_set_popdown_strings(GTK_COMBO(globalState.mscCombo),
				      tokenItems);

	return TRUE;
}

BOOL MSCFormat(pFormatMuscleInfo formatInfo)
{
	MSCInitTokenParams initParams;
	int i;
	MSC_RV rv;
	MSCString mem;
	MSCString tries;
	MSCPUChar8 newAdmin;
	MSCPUChar8 userPIN;
	MSCPUChar8 userUnblock;
	MSCUChar8 numTries;
	MSCUChar8 userTransportKey[16];
	MSCULong32 userKeyLength = 16;
	MSCULong32 memory;

	pthread_t pthThread;

	if (!textToBytes(gtk_entry_get_text(formatInfo->transportKeyEntry),
			 userTransportKey, &userKeyLength)) {
		DEBUG("Transport key error\n");
		return FALSE;
	}

	mem = gtk_entry_get_text(formatInfo->memoryEntry);
	if (mem[0] == '\0') {
		messageBox("Please enter memory size.");
		return FALSE;
	}
	for (i = 0; i < strlen(mem); i++) {
		if (!isdigit(mem[i])) {
			messageBox("Error parsing memory: \"%s\" at %d.",
				   mem, i);
			return FALSE;
		}
	}
	memory = atoi(mem);

	tries = gtk_entry_get_text(formatInfo->pinTriesEntry);
	if (tries[0] == '\0') {
		messageBox("Please enter PIN tries.");
		return FALSE;
	}
	for (i = 0; i < strlen(tries); i++) {
		if (!isdigit(tries[i])) {
			messageBox
			    ("Error parsing number of tries: \"%s\" at %d.",
			     tries, i);
			return FALSE;
		}
	}
	numTries = atoi(tries);

	newAdmin = gtk_entry_get_text(formatInfo->authPinEntry);
	if (newAdmin[0] == '\0') {
		messageBox("Please enter admin PIN.");
		return FALSE;
	}
	if (strlen(newAdmin) < 4) {
		DEBUG("Length error, admin\n");
		messageBox("Error: Admin PIN to short, minimum 4 characters.");
		return FALSE;
	}
	userPIN = gtk_entry_get_text(formatInfo->userPinEntry);
	if (userPIN[0] == '\0') {
		messageBox("Please enter user PIN.");
		return FALSE;
	}
	if (strlen(userPIN) < 4) {
		DEBUG("Length error, user\n");
		messageBox("Error: User PIN to short, minimum 4 characters.");
		return FALSE;
	}
	userUnblock = gtk_entry_get_text(formatInfo->pinUnblockEntry);
	if (userUnblock[0] == '\0') {
		messageBox("Please enter unblock PIN.");
		return FALSE;
	}
	if (strlen(userUnblock) < 4) {
		DEBUG("Length error, unblock\n");
		messageBox
		    ("Error: Unblock PIN to short, minimum 4 characters.");
		return FALSE;
	}

	initParams.transportBehavior = MSC_INIT_USE_KEY;
	memcpy(initParams.transportKey, userTransportKey, userKeyLength);
	initParams.transportKeyLen = userKeyLength;

	memcpy(initParams.newTransportKey, newAdmin, strlen(newAdmin));
	initParams.newTransportKeyLen = strlen(newAdmin);

	memcpy(initParams.defaultCHV, userPIN, strlen(userPIN));
	initParams.defaultCHVLen = strlen(userPIN);

	memcpy(initParams.defaultCHVUnblock, userUnblock, strlen(userUnblock));
	initParams.defaultCHVUnblockSize = strlen(userUnblock);

	initParams.defaultCHVTries = (MSCUChar8) numTries;
	initParams.defaultCHVUnblockTries = 3;
	initParams.objectMemory = memory;

	if (globalState.p11Initialized == TRUE) {
		C_Finalize(NULL);
		globalState.p11Initialized = FALSE;
	}

	globalState.pConnection = malloc(sizeof(MSCTokenConnection));

	//  DEBUG("TokenSize: %d\tTokenList addr: %x\n", globalState.tokenSize,
	//    globalState.tokenList);

	//  DEBUG("Reader: %d\n", globalState.mscSelection);
	rv = MSCEstablishConnection(&globalState.
				    tokenList[globalState.mscSelection],
				    MSC_SHARE_DIRECT, 0, 0,
				    globalState.pConnection);
	if (rv != MSC_SUCCESS) {
		free(globalState.pConnection);
		globalState.pConnection = 0;
		free(globalState.tokenList);
		globalState.tokenList = 0;
		display_main
		    ("Error: EstablishConnection Failed! Could not format token.");
		DEBUG("ERR: EstablishConnection Failed!");
		return FALSE;
	}

	DELETE();
	DISPLAY("Formatting token [");
	FLUSH();

	globalState.quitStat = FALSE;
	rv = pthread_create(&pthThread, NULL, (void *)&printStatStars, 0);

	rv = MSCWriteFramework(globalState.pConnection, &initParams);
	MSCReleaseConnection(globalState.pConnection, MSC_RESET_TOKEN);

	free(globalState.pConnection);
	globalState.pConnection = 0;
	free(globalState.tokenList);
	globalState.tokenList = 0;

	globalState.quitStat = TRUE;
	while (globalState.quitStat == TRUE)
		usleep(1000*10);

	if (rv != MSC_SUCCESS) {
		DISPLAY("FAILED");
		DEBUG("Failed (%s)\n", msc_error(rv));
		return FALSE;
	} else {
		DISPLAY("SUCCESS");
	}

	return TRUE;
}

#endif
